/********************************************************************************
* Program to demonstrate the usage of the FFTW-library to calculate derivatives	*
*										*
* author: Johannes Luelff							*
*										*
********************************************************************************/


#include <iostream>
#include <fstream>
#include <cmath>
#include "fftw3.h"

using namespace std;


//constants
const double PI = 3.141592653;	
const int N = 256;
const double L = 2*PI;


//Function; x ranges from 0 to L
double func(double x) {
	//return sin(3*x);
	return exp(-4*pow(x-L/2.0, 2));
}

//1st derivative of Function; x ranges from 0 to L
double dxfunc(double x) {
	//return 3*cos(3*x);
	return -8*(x-L/2.0)*exp(-4*pow(x-L/2.0, 2));
}

//2nd derivative of Function; x ranges from 0 to L
double dx2func(double x) {
	//return -9*sin(3*x);
	return -8*exp(-pow(x-L/2.0, 2)) + 64*(x-L/2.0)*(x-L/2.0)*exp(-pow(x-L/2.0, 2));
}


//calculates a = a*b in complex
void cMult(double *a, double *b) {
	double re, im;
	re = a[0]*b[0] - a[1]*b[1];
	im = a[1]*b[0] + a[0]*b[1];
	a[0] = re;
	a[1] = im;
}

//saves the array d and the difference to the exact solution in ASCII to the stream f
void saveToStream(double *d, ofstream &f, double (*exact)(double)) {
	for(int i=0; i<N; i++) 
		f << L*i/double(N) << " " << d[i] << " " << fabs(d[i]-exact(L*i/double(N))) << endl;
}

//gererates 1st derivative of array d and exact solution
void generateAndSaveFiniteDifferences(double *d, ofstream &f, double (*exact)(double)) {	
	double fd;
	double h = L / double(N);
	for(int i=0; i<N; i++) {
		if(i==0)
			fd = (d[1]-d[N-1]) / h;
		else
			if(i==N-1)
				fd = (d[0] - d[N-2]) / (2*h);
			else
				fd = (d[i+1] - d[i-1]) / (2*h);
		f << L*i/double(N) << " " << fd << " " << fabs(fd-exact(L*i/double(N))) << endl;
	}
}


int main() {

	//files for output
	ofstream f("f.txt");
	ofstream dxf("dxf.txt");
	ofstream dx2f("dx2f.txt");
	ofstream finitef("finitef.txt");

	//arrays holding...	      (fftw_complex is just a double[2])
	double *real;		//... the function
	fftw_complex *complex;	//... the FT of the function
	fftw_complex *k, *k2;	//... the k-vectors	

	//use optimized fftw routines to allocate memory
	real 	= (double*)	 fftw_malloc(sizeof(double)*N);
	complex = (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*(N/2+1));
	k 	= (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*(N/2+1));
	k2	= (fftw_complex*)fftw_malloc(sizeof(fftw_complex)*(N/2+1));

	//initialize and save the function to derive
	for(int i=0; i<N; i++) 
		real[i] = func(i/double(N) * L);
	saveToStream(real, f, func);

	//initialize the complex k-vectors
	for(int i=0; i<N/2+1; i++) {
		//i*k for 1st derivative
		k[i][0] = 0;	
		k[i][1] = 2*PI*i/L;
		//-k**2 for 2nd derivative
		k2[i][0] = -pow(2*PI*i/L, 2);
		k2[i][1] = 0;
	}

	//FFTW-Interna; make so-called plans of the transforms you are about to perform
	//plan for converting the array "real" which holts N points from real to complex space
	fftw_plan planR2C = fftw_plan_dft_r2c_1d(N, real, complex, FFTW_ESTIMATE);
	//plan for converting the array "complex" which holts N points from complex to real space
	fftw_plan planC2R = fftw_plan_dft_c2r_1d(N, complex, real, FFTW_ESTIMATE);


	//execute the specified plan, namely convert to fourier space
	fftw_execute(planR2C);
	//multiply the fourier transform with the k-vector
	for(int i=0; i<N/2+1; i++)
		cMult(complex[i], k[i]);
	//convert back to spatial domain
	fftw_execute(planC2R);
	//normalize the vector, b/c fftw calculates _unnormalized_ FT
	for(int i=0; i<N; i++)
		real[i] /= double(N);
	//save the result, namely the 1st derivative
	saveToStream(real, dxf, dxfunc);


	//same for 2nd derivative
	for(int i=0; i<N; i++) 
		real[i] = func(i/double(N) * L);
	fftw_execute(planR2C);
	for(int i=0; i<N/2+1; i++)
		cMult(complex[i], k2[i]);
	fftw_execute(planC2R);
	for(int i=0; i<N; i++)
		real[i] /= double(N);
	saveToStream(real, dx2f, dx2func);


	//calculate derivative via finite differences
	for(int i=0; i<N; i++) 
		real[i] = func(i/double(N) * L);
	generateAndSaveFiniteDifferences(real, finitef, dxfunc);


	//clean up FFTW-related stuff after we're ready
	fftw_destroy_plan(planR2C);
	fftw_destroy_plan(planC2R);

	fftw_free(real);
	fftw_free(complex);
	fftw_free(k);
	fftw_free(k2);	
}
